#
# Edits is a list of edits for undo/redo and also to know what to sync to o/mero.
#
# Edits.pos points always to the "current" edit.
# If the current edit is not closed, further edits are added if feasible by modifying
# the current edit (inserting even more past the current insert, etc.)
# If the current edit is closed, further edits are added just past it (a newly open edit).
# Events synced have the flag synced set to true.
# Events past the current edit were undone. Their synced flag reports if they have been
# seen by o/mero (not yet undone in o/mero) or not.
# If, when adding a new edit (not combined in the current) there are edits pending to be synced,
# the module reports failure. The caller is expected to sync the changes and try again.
# But note that undos are pending to be synced when they *have* the synced flag (o/mero did
# see them, but we have yet to undo that).
#
# In general, only the last edit may be synced or not. The history looks like
# 	closed	synced	ins blah
# 	closed	synced	ins y
# ->	closed	synced	del x	(current)
#	closed	synced	ins z
#			del a
# or
# 	closed	synced	ins blah
# 	closed		ins y
# ->	closed		del x	(current) 
#	closed		ins z
#			del a
#
# The former has some undos (not undone in o/mero), the later has undos (undone
# in o/mero). Without undos, it is usual to have something like:
#
# 	closed	synced	ins blah
# 	closed	synced	ins y
# ->			del x	(current) 
#

implement Tundo;
include "sys.m";
	sys: Sys;
	fprint, sprint: import sys;
include "error.m";
	err: Error;
	checkload, stderr, panic, kill, error: import err;
include "string.m";
include "tblks.m";
	tblks: Tblks;
	fixposins, fixposdel: import tblks;
include "lists.m";
	lists: Lists;
	concat: import lists;
include "tundo.m";

debug := 2;

init(sysm: Sys, e: Error, l: Lists, t: Tblks, dbg: int)
{
	sys = sysm;
	err = e;
	tblks = t;
	lists = l;
	debug = dbg;
}

Edit.name(edit: self ref Edit): string
{
	pick edp := edit {
	Ins =>
		return "ins";
	Del =>
		return "del";
	}
	return nil;
}

reverse(ed: ref Edit): ref Edit
{
	pick edp := ed {
	Ins =>
		return ref Edit.Del(1, 1, edp.bundle, edp.pos, edp.s);
	Del =>
		return ref Edit.Ins(1, 1, edp.bundle, edp.pos, edp.s);
	}
	return nil;
}

nulledit(ed: ref Edit): int
{
	return len ed.s == 0;
}


Edits.new(): ref Edits
{
	ed0 := ref Edit.Del(1, 1, 0, 0, "");	# fake null del at pos 0 (marker).
	edits := ref Edits;
	edits.e = array[1] of ref Edit;
	edits.e[0] = ed0;
	edits.pos = edits.cpos = 0;
	edits.gen++;
	return edits;
}

grow(edits: ref Edits, n: int)
{
	ne := array[len edits.e+n] of ref Edit;
	ne[0:] = edits.e;
	for(i := len edits.e; i < len ne; i++)
		ne[i] = nil;
	edits.e = ne;
}

editsend(edits: ref Edits): int
{
	for(e := edits.pos; e < len edits.e && edits.e[e] != nil; e++)
		;
	return e;
}

# reports changes needed to sync o/mero with us.
# updates their synced flags if sync != 0
changes(edits: ref Edits, sync: int): list of ref Edit
{
	nl : list of ref Edit;

	# old edits not yet synced
	for(i := 1; i <= edits.pos && edits.e[i] != nil; i++){
		if(!nulledit(edits.e[i]) && !edits.e[i].synced)
			nl = concat(nl, list of {edits.e[i]});
		if(sync)
			edits.e[i].synced = edits.e[i].closed = 1;
	}

	# undos not undone in o/mero in reverse order
	for(i = edits.pos+1; i < len edits.e && edits.e[i] != nil; i++){
		if(!nulledit(edits.e[i]) && edits.e[i].synced)
			nl = reverse(edits.e[i])::nl;
		if(sync){
			edits.e[i].synced = 0;
			edits.e[i].closed = 1;
		}
	}
	return nl;
}

# inserts neds starting at pos, and returns and prunes old edits from there on.
xchgedits(edits: ref Edits, neds: list of ref Edit, pos: int): array of ref Edit
{
	e := editsend(edits);
	l := len neds;
	if(e + l > len edits.e)
		grow(edits, l);	# more than needed.
	oe := array[e-pos] of ref Edit;
	for(i := pos; i < e; i++){
		oe[i-pos] = edits.e[i];
		edits.e[i] = nil;
	}
	for(; neds != nil; neds = tl neds)
		edits.e[pos++] = hd neds;
	return oe;
}

revl(l: list of ref Edit): list of ref Edit
{
	nl: list of ref Edit;
	for(; l != nil; l = tl l)
		nl = hd l::nl;
	return nl;
}

Edits.undo(edits: self ref Edits): list of ref Edit
{
	if(edits.pos == 0)		# null 0th edit is always kept
		return nil;		# to avoid special cases.
	if(edits.e[edits.pos] == nil)
		panic("edits undo bug");
	edits.e[edits.pos].closed = 1;
	l := list of {reverse(edits.e[edits.pos])};
	for(edits.pos--; edits.pos > 0; edits.pos--)
		if(edits.e[edits.pos].bundle != edits.e[edits.pos+1].bundle)
			break;
		else
			l = reverse(edits.e[edits.pos])::l;
	return revl(l);
}

Edits.redo(edits: self ref Edits): list of ref Edit
{
	if(edits.pos+1 >= len edits.e || edits.e[edits.pos+1] == nil)
		return nil;
	l := list of {edits.e[++edits.pos]};
	while(edits.pos+1 < len edits.e && edits.e[edits.pos+1] != nil)
		if(edits.e[edits.pos].bundle != edits.e[edits.pos+1].bundle)
			break;
		else
			l = edits.e[++edits.pos]::l;
	return revl(l);
}

# About to add more edits. Ensure edits.e[edits.pos] is ok to ins/del.
# After mkpos, edits.pos is either not closed or it is a null entry.
# Further undone edits are pruned, to keep just a single edit history.
# If there are pending syncs, we report failure to let the caller sync and retry.
Edits.mkpos(edits: self ref Edits): int
{
	npos := edits.pos;
	if(edits.e[npos] != nil && edits.e[npos].closed)
		npos++;

	for(i := npos; i < len edits.e && edits.e[i] != nil; i++)
		if(edits.e[i].synced)
			return -1;
	if(!edits.e[npos-1].synced)
		return -1;	

	for(i = npos; i < len edits.e && edits.e[i] != nil; i++)
		if(edits.e[i].closed || i > npos)
			edits.e[i] = nil;
	if(npos >= len edits.e)
		grow(edits, 3);
	edits.pos = npos;
	return 0;
}

Edits.ins(edits: self ref Edits, s: string, pos: int): int
{
	if(len s == 0){
		fprint(stderr, "edits: ins: null string\n");
		return 0;
	}
	if(edits.mkpos() < 0)
		return -1;
	if(edits.e[edits.pos] == nil)			# can't fail later
		edits.e[edits.pos] = ref Edit.Ins(0, 0, edits.gen++, pos, "");
	pick edp := edits.e[edits.pos] {
	Ins =>
		if(pos < edp.pos || pos > edp.pos + len edp.s){
			edits.e[edits.pos].closed = 1;
			return -1;
		}
		l := len edp.s;
		pos -= edp.pos;
		if(len s == 1 && pos == l)
			edp.s[l] = s[0];
		else {
			ns := edp.s[0:pos] + s;
			if(pos < l)
				ns += edp.s[pos:];
			edp.s = ns;
		}
	Del =>
		edits.e[edits.pos].closed = 1;
		return -1;
	}
	return 0;
}

Edits.del(edits: self ref Edits, s: string, pos: int): int
{
	if(len s == 0){
		fprint(stderr, "edits: del: null string\n");
		return 0;
	}
	if(edits.mkpos() < 0)
		return -1;
	if(edits.e[edits.pos] == nil)			# can't fail later
		edits.e[edits.pos] = ref Edit.Del(0, 0, edits.gen++, pos, "");
	pick edp := edits.e[edits.pos] {
	Ins =>
		epos := pos + len s;
		if(pos < edp.pos || epos > edp.pos + len edp.s){
			edits.e[edits.pos].closed = 1;
			return -1;
		}
		pos -= edp.pos;
		epos-= edp.pos;
		ns := edp.s[0:pos];
		if(epos < len edp.s)
			ns += edp.s[epos:];
		edp.s = ns;
	Del =>
		if(pos + len s == edp.pos){
			edp.pos = pos;
			edp.s = s + edp.s;
		} else if(pos == edp.pos + len edp.s)
			edp.s += s;
		else {
			edits.e[edits.pos].closed = 1;
			return -1;
		}
	}
	return 0;
}

# xedit combines external edits wrt our possibly ongoing edits.
# This is the core of the cooperative editing. It fixes the external
# edit wrt local not synced edits, pretending it was made before the
# them. It is inserted in place.
# The routine returns changes to be made to text/frame to apply the
# external edit.

Edits.xedit(edits: self ref Edits, xed: ref Edit): list of ref Edit
{
	if(debug>1){
		fprint(stderr, "before %s %d [%s]\n", xed.name(), xed.pos, xed.s);
		edits.dump();
	}

	# 1. fix xed wrt local edits
	eds := changes(edits, 0);
	xeds := list of {xed};
	for(l := eds; l != nil; l = tl l)
		xeds = fixeditsafter(xeds, hd l);

	# 2. put fixed xeds before our ongoing edits, as synced,
	# removing our old ongoing edits.
	for(lpos := edits.pos; lpos > 0; lpos--)
		if(edits.e[lpos] != nil && edits.e[lpos].synced)
			break;
	lpos++;
	leds := xchgedits(edits, xeds, lpos);
	edits.pos += len xeds;

	# 3. fix ongoing edits wrt xeds and add them back in place
	e := editsend(edits);
	for(i := 0; i < len leds; i++)
		for(xl := xeds; xl != nil; xl = tl xl){
			neds := fixeditafter(leds[i], hd xl);
			xchgedits(edits, neds, e);
			e += len neds;
		}
	if(debug>1){
		fprint(stderr, "after: (%d leds)\n", len leds);
		edits.dump();
		dumpedits("with xeds:", xeds);
	}
	return xeds;
}

fixeditafter(ed: ref Edit, ned: ref Edit): list of ref Edit
{
	pick edp := ed {
	Ins =>
		pick nedp := ned {
		Ins =>
			ed.pos = fixposins(ed.pos, ned.pos, len ned.s);
		Del =>
			ed.pos = fixposdel(ed.pos, ned.pos, len ned.s);
		}
		return list of {ed};
	Del =>
		if(ned.pos >= ed.pos + len ed.s)
			return list of {ed};
		pick nedp := ned {
		Ins =>
			if(ned.pos < ed.pos){
				ed.pos = fixposins(ed.pos, ned.pos, len ned.s);
				return list of {ed};
			} else {
				# ned inserted in the middle of del: split it.
				ed1 := ref Edit.Del(ed.synced, 1, ed.bundle, ed.pos, ed.s);
				ed1.s = ed1.s[0:ned.pos - ed.pos];

				ed2 := ref Edit.Del(ed.synced, 1, ed.bundle, ed.pos, ed.s);
				ed2.pos = ned.pos + len ned.s;
				ed2.s = ed2.s[ned.pos - ed.pos:];
				return list of {ed1, ed2};
			}
		Del =>
			if(ned.pos + len ned.s <= ed.pos){
				ed.pos = fixposdel(ed.pos, ned.pos, len ned.s);
				return list of {ed};
			}
			# remove from ed the part already deleted by ned.
			lpos := ned.pos;
			if(lpos < ed.pos)
				lpos = ed.pos;
			rpos := ned.pos + len ned.s;
			if(rpos  > ed.pos + len ed.s)
				rpos = ed.pos + len ed.s;
			loff := lpos - ed.pos;
			roff := rpos - ed.pos;
			ns := ed.s[0:loff];
			if(roff < len ed.s)
				ns += ed.s[roff:];
			ed.s = ns;
			ed.pos = fixposdel(ed.pos, ned.pos, len ned.s);
			return list of {ed};
		}
	}
}

fixeditsafter(xeds: list of ref Edit, ed: ref Edit): list of ref Edit
{
	nxeds: list of ref Edit;
	for(; xeds != nil; xeds = tl xeds)
		nxeds = concat(nxeds, fixeditafter(hd xeds, ed));
	return nxeds;	
}


Edits.sync(edits: self ref Edits): list of ref Edit
{
	return changes(edits, 1);
}

dtxt(s: string): string
{
	if(len s> 35)
		s = s[0:15] + " ... " + s[len s - 15:];
	ns := "";
	for(i := 0; i < len s; i++)
		if(s[i] == '\n')
			ns += "\\n";
		else
			ns[len ns] = s[i];
	return ns;
}

Edit.text(e: self ref Edit): string
{
	if(e == nil)
		return "<nil>";
	s := "  ";
	if(e.synced)
		s[0] = 's';
	if(e.closed)
		s[1] = 'c';
	s += sprint("%s pos %d\t'%s'", e.name(), e.pos, dtxt(e.s));
	return s;
}

Edits.dump(edits: self ref Edits)
{
	fprint(stderr, "%d edits (pos %d cpos %d)\n", len edits.e, edits.pos, edits.cpos);
	for(i := 0; i < len edits.e && edits.e[i] != nil; i++){
		e := edits.e[i];
		s := "   ";
		if(i == edits.pos)
			s = " ->";
		if(i == edits.cpos)
			s[0] = 'f';
		s = sprint("[%02d] %s\t%s", i, s, e.text());
		fprint(stderr, "%s\n", s);
	}
	fprint(stderr, "\n");
}

dumpedits(s: string, l: list of ref Edit)
{
	fprint(stderr,"%s:\n", s);
	for(; l != nil; l = tl l)
		fprint(stderr, "\t%s\n", (hd l).text());
}
