# Generic panel coordination
# One process (per tree) coordinates operations within the tree, so that
# there are no races.
# A tree operation operates on a subtree, identified by its root panel.
# The operation blocks the subtree while it is being executed by a helper process.
# A operation can proceed if
#	1. no inner panel of its subtree is blocked by a previous operation
#	2. the root of its subtree is not within a blocked subtree
# When it cannot proceed, it is queued along with the panel causing it to block
# When an operation completes, we process its queued operations one by one
# unless they block again.

# Tree operations that may take a long time are performed by auxiliary
# processes, represented by Xops (to be cached and reused).
# Channels of type ref Panel are also cached by means of get/putpchan.

implement Wtree;
include "mods.m";
mods, debug, win, tree: import dat;

Menu: import menus;
Ptag, Pline, Pedit, Pdead, Pdirty, Pshown, Predraw, Phide, Qatom,
Pbusy, Pdirties, Pmore, Playout, Qrow, Qcol, nth, Pinsist,
intag, Panel: import wpanel;
Cpointer, Tagwid, Taght, Inset, setcursor, drawtag, lastxy, terminate,
getfont, readsnarf, cookclick, Arrow, Drag, Resize: import gui;
dirname, basename: import names;

# The following code is support to reuse processes
# heavily spawned by the tree.
# Should be replaced by using xproc(2) in the future.

# tree operations performed by auxiliary processes
Xop: adt {
	f:	ref fn(x: ref Xop);
	t:	ref Tree;
	rp:	ref Panel;
	op:	ref Treeop;
	dc:	chan of ref Panel;
	rc:	chan of ref Panel;
	d:	ref Dir;			# for update tree rec.
};

xc: chan of ref Xop;

pchans: list of chan of ref Panel;
getpchan(): chan of ref Panel
{
	c: chan of ref Panel;
	if(pchans != nil){
		c = hd pchans;
		pchans = tl pchans;
	} else
		c = chan of ref Panel;
	return c;
}

putpchan(c: chan of ref Panel)
{
	pchans = c::pchans;
}

init(d: Livedat)
{
	dat = d;
	initmods();
	xc = chan of ref Xop;
	spawn xctlproc();
}

# This should now use xproc(2)
xctlproc()
{
	tprocs: list of chan of ref Xop;
	tprocs = nil;
	nprocs := 0;
	idlec := chan of chan of ref Xop;
	for(;;){
		alt {
		x := <- xc =>
			tpc: chan of ref Xop;
			if(tprocs != nil){
				tpc = hd tprocs;
				tprocs = tl tprocs;
			} else {
				tpc = chan of ref Xop;
				spawn xproc(tpc, idlec);
				if(debug['P'])
					fprint(stderr, "%d tree procs\n", nprocs);
			}
			tpc <-= x;
		tpc := <- idlec =>
			tprocs = tpc::tprocs;
		}
	}
}

xproc(tpc: chan of ref Xop, idlec: chan of chan of ref Xop)
{
	for(;;){
		x := <-tpc;
		x.f(x);
		if(x.rc != nil)
			x.rc <-= x.rp;
		if(x.dc != nil)
			x.dc <-= x.rp;
		idlec <-= tpc;
	}
}


Tree.path(t: self ref Tree, p: ref Panel): string
{
	l := len t.slash.path;
	path := p.path[l:];
	if(path == "")
		path = "/";
	return path;
}

notshown(p: ref Panel, prunep: ref Panel)
{
	if(p == prunep)
		return;
	p.flags &= ~Pshown;
	if(p.rowcol)
		for(i := 0; i < len p.child; i++)
			notshown(p.child[i], prunep);
}



showtree(p: ref Panel, force: int)
{
	mustdraw := p.flags&Predraw;
	p.flags &= ~Predraw;
	if(p.rect.dx() <= Inset || p.rect.dy() <= Inset || (p.flags&Phide)){
		notshown(p, nil);
		return;
	}
	p.flags |= Pshown;
	force |= ! p.rect.eq(p.orect);
	if(p.rowcol != Qatom){
		for(i := 0; i < len p.child; i++)
			showtree(p.child[i], force);
		p.draw();
	} else
		if(force || mustdraw){
			if(p.debug('L'))
			fprint(stderr, "show: %s\t[%d %d %d %d] drw%d frz%d dp=%d\n", p.text(),
				p.rect.min.x, p.rect.min.y, p.rect.max.x, p.rect.max.y, mustdraw, force, p.depth);
			p.draw();	# must draw tag as well
			if(p.debug('D'))
				sys->sleep(2000);
		}
}

showtreeproc(x: ref Xop)
{
	force := x.t == nil;	# kludge.
	showtree(x.rp, force);
}

orderchildren(p: ref Panel)
{
	ochild := p.child;
	n := len ochild;
	if(n < len p.order)
		n = len p.order;
	nchild := array[n] of ref Panel;
	pos := 0;
	for(ol := p.order; ol != nil; ol = tl ol)
		for(i := 0; i < len ochild; i++)
			if(ochild[i] != nil && hd ol == ochild[i].name){
				nchild[pos++] = ochild[i];
				ochild[i] = nil;
				break;
			}
	for(i = 0; i < len ochild; i++)
		if(ochild[i] != nil)
			nchild[pos++] = ochild[i];
	p.child = nchild[0:pos];
}

packchildren(p: ref Panel)
{
	i,j : int;
	for(i = j = 0; j < len p.child; j++)
		if((p.child[i] = p.child[j]) != nil)
			i++;
	p.child = p.child[0:i];
}

unescape(s: string): string
{
	for(i := 0; i < len s; i++)
		if(s[i] == 1)
			s[i] = '\n';
	return s;
}

updatetree(p: ref Panel, newname: string, vers: int,
	ctls: string, data: array of byte, edits: string)
{
	if(newname != nil){
		dp := p;
		p = Panel.new(newname, dp);
		if(p == nil)
			return;
		p.flags |= Predraw;
		orderchildren(dp);
		dp.flags |= Predraw;
	}

	if(data != nil){
		if(p.debug('P')){
			dd := data;
			if(len dd > 10)
				dd = dd[0:10];
			fprint(stderr, "o/live: %s: updatedata vers %d [%s...]\n",
				p.path, p.vers, string dd);
		}
		p.update(data);
	}
	if(ctls != nil){
		p.flags |= Predraw;
		(nil, cl) := tokenize(ctls, "\n");
		for(; cl != nil; cl = tl cl)
			p.ctl(hd cl);
	}
	if(edits != nil){
		(nil, cl) := tokenize(edits, "\n");
		for(; cl != nil; cl = tl cl)
			p.event("edit" + unescape(hd cl));
	}
	p.vers = vers;
}

Tree.update(t: self ref Tree, path: string, vers: int,
	ctls: string, data: array of byte, edits: string)
{
	if(debug['P'] > 1){
		cs := ds := es := "";
		if(ctls != nil)
			cs = " ctls";
		if(data != nil)
			ds = " data";
		if(edits != nil)
			es = " edits";
		fprint(stderr, "o/live: updatetree %s%s%s%s\n", path, cs, ds, es);
	}
	rc := getpchan();
	t.opc <-= ref Treeop.Update(rc, path, vers, ctls, data, edits);
	<-rc;
	putpchan(rc);
}

deltree(p: ref Panel)
{
	p.flags |= Pdead;
	for(i := 0; i < len p.child; i++)
		deltree(p.child[i]);
	p.term();
}

closetree(p: ref Panel)
{
	p.flags |= Pdead;
	dp := p.parent;
	if(dp == nil)
		terminate("Root panel removed");
	for(i := 0; i < len dp.child; i++)
		if(dp.child[i] == p){
			dp.child[i] = nil;
			break;
		}
	if(i == len dp.child)
		fprint(stderr, "o/live: closetree: not linked\n");
	packchildren(dp);
	deltree(p);
}

Tree.close(t: self ref Tree, path: string)
{
	if(debug['P'] > 1)
		fprint(stderr, "o/live: closetree %s\n", path);
	rc := getpchan();
	t.opc <-= ref Treeop.Close(rc, path);
	<-rc;
	putpchan(rc);
}

mousetreeproc(x: ref Xop)
{
	pick op := x.op {
	Mouse =>
		x.rp.mouse(op.m, op.mc);
	}
}

Tree.mouse(t: self ref Tree, m: ref Cpointer, mc: chan of ref Cpointer): ref Panel
{
	rc := getpchan();
	t.opc <-= ref Treeop.Mouse(rc, nil, m, mc);
	p := <-rc;
	putpchan(rc);
	return p;
}

kbdtreeproc(x: ref Xop)
{
	pick op := x.op {
	Kbd =>
		x.rp.kbd(op.r);
	}
}

Tree.kbd(t: self ref Tree, r: int): ref Panel
{
	rc := getpchan();
	t.opc <-= ref Treeop.Kbd(rc, nil, r);
	p := <-rc;
	putpchan(rc);
	return p;
}

tabs(n: int): string
{
	s := "";
	for(i := 0; i < n; i++)
		s += "  ";
	return s;
}

dumptree(p: ref Panel, d: int)
{

	fprint(stderr, "%s%s\n", tabs(d), p.text());
	if(p.rowcol != Qatom)
		for(i := 0; i < len p.child; i++)
				dumptree(p.child[i], d+1);
}

checkpath(path: string): string
{
	if(path == nil)
		path = "/";
	if(path[0] != '/')
		panic("Tree: relative path");
	return path;
}

Tree.dump(t: self ref Tree, path: string)
{
	path = checkpath(path);
	rc := getpchan();
	t.opc <-= ref Treeop.Dump(rc, path);
	<-rc;
	putpchan(rc);
}

# locates a point, stops if any from "/" to the panel is busy
ptwalktree(p: ref Panel, pt: Point, atomok: int): ref Panel
{
	if(p.flags&Pdead)	# let the caller check it out.
		return p;
	if(p.flags&Pbusy)
		return nil;
	if(pt.in(p.rect) && p.rowcol != Qatom)
		for(i := 0; i < len p.child; i++){
			np := p.child[i];
			if(np.flags&Phide)
				continue;
			if(pt.in(np.rect) && (atomok || np.rowcol != Qatom))
				return ptwalktree(np, pt, atomok);
		}
	return p;
}

Tree.ptwalk(t: self ref Tree, pt: Point, atomok: int): ref Panel
{
	rc := getpchan();
	t.opc <-= ref Treeop.Ptwalk(rc, "/", pt, atomok);
	p := <-rc;
	putpchan(rc);
	return p;
}

walktree(fp: ref Panel, elems: list of string): ref Panel
{
	if(len elems == 0)
		return fp;
	if(hd elems == "/")
		return walktree(fp, tl elems);
	for(i := 0; i < len fp.child; i++)
		if(fp.child[i].name == hd elems)
			return walktree(fp.child[i], tl elems);
	return nil;
}

Tree.walk(t: self ref Tree, path: string): ref Panel
{
	path = checkpath(path);
	rc := getpchan();
	t.opc <-= ref Treeop.Walk(rc, path);
	p := <-rc;
	putpchan(rc);
	if(p == nil)
		fprint(stderr, "o/live: Tree.walk: %s: not found\n", path);
	return p;
}

# It would be more efficient to compute this as we change the tree
# But it's more simple to get it correct this way.
tagtree(fp: ref Panel, hidden: int): (int, int)
{
	dirties := (fp.flags&Pdirty);
	more := (fp.flags&Phide);
	hidden |= (fp.flags&Phide);
	fp.nshown = 0;
	for(i := 0; i < len fp.child; i++){
		(cd, cm) := tagtree(fp.child[i], hidden);
		dirties |= cd;
		more |= cm;
		if(!(fp.child[i].flags&Phide))
			fp.nshown++;
	}
	if(fp.rowcol){
		if(dirties)
			fp.flags |= Pdirties;
		else
			fp.flags &= ~Pdirties;
		if(more)
			fp.flags |= Pmore;
		else
			fp.flags &= ~Pmore;
	}
	if(!hidden && (fp.flags&Pshown) && (fp.flags&Ptag))
		drawtag(fp);
	return (dirties, more);
}

Tree.tags(t: self ref Tree, path: string)
{
	path = checkpath(path);
	rc := getpchan();
	t.opc <-= ref Treeop.Tags(rc, path);
	<-rc;
	putpchan(rc);
}

Tree.size(t: self ref Tree, path: string, op: int)
{
	path = checkpath(path);
	rc := getpchan();
	t.opc <-= ref Treeop.Size(rc, path, op);
	<-rc;
	putpchan(rc);
}

Tree.layout(t: self ref Tree, path: string, auto: int)
{
	if(path == nil)
		path = t.path(t.dslash);
	path = checkpath(path);
	rc := getpchan();
	t.opc <-= ref Treeop.Layout(rc, path, auto);
	<-rc;
	putpchan(rc);
}

paneltop(p: ref Panel): ref Panel
{
	for(;;){
		if(p.parent == nil)
			return p;
		if(p.flags&Playout)
			return p;
		p = p.parent;
	}
}

unblock(l: list of string, p: string): list of string
{
	nl: list of string;
	for(; l != nil; l = tl l){
		if(hd l != p)
			nl = hd l::nl;
	}
	return nl;
}

isblocked(path: string, bl: list of string): int
{
	for(; bl != nil; bl = tl bl)
		if(path == hd bl || isprefix(path, hd bl) || isprefix(hd bl, path))
			return 1;
	return 0;
}

release(hl: list of ref Treeop, bl: list of string): (list of ref Treeop, list of ref Treeop)
{
	nhl: list of ref Treeop;
	nrl: list of ref Treeop;

	for(; hl != nil; hl = tl hl){
		if(isblocked((hd hl).path, bl))
			nhl = hd hl::nhl;
		else
			nrl = hd hl::nrl;
	}
	return (nrl, nhl);
}

nofocus(t: ref Tree, p: ref Panel)
{
	t.donec <-= p;
}

showchildren(p: ref Panel, first: int, last: int, only: ref Panel)
{
	this := 0;
	for(i := 0; i < len p.child; i++){
		np := p.child[i];
		if(first <= this && this < last || np == only){
			if(np.flags&Phide){
				np.ctl("show");
				np.fsctl("show", 1);
			}
		} else
			if(!(np.flags&Phide)){
				np.ctl("hide");
				np.fsctl("hide", 1);
			}
		this++;
	}
}

sizetree(op: ref Treeop.Size, rp: ref Panel)
{
	case op.op {
	More =>
		if(rp.nshown >= len rp.child)
			showchildren(rp, 0, 1, nil);
		else
			showchildren(rp, 0, rp.nshown+1, nil);
	Full =>
		showchildren(rp, 0, len rp.child, nil);
	Max =>
		if((fp := rp.parent) != nil && fp != rp)
			showchildren(fp, 0, 0, rp);
	}
}

treeproc(t: ref Tree)
{
	blocked: list of string;	# list of blocked paths for subtrees
	onhold: list of ref Treeop;	# ops waiting because of blocked subtrees
	released: list of ref Treeop;	# were on hold, attend before t.opc
	blkd: int;
	t.focus = nil;
Loop:
	for(;;){
		# 1. Get the event and pay attention to busy panels.
		#
		o: ref Treeop;
		if(released != nil){
			o = hd released;
			released = tl released;
		}
		if(o == nil){
			alt {
			o = <-t.opc =>
				;
			p := <-t.donec =>
				p.flags &= ~Pbusy;
				blocked = unblock(blocked, p.path);
				(released, onhold) = release(onhold, blocked);
				continue Loop;
			}
		}

		# Mouse ops may draw menus and bars that require the
		# screen to be quiet. Other operations do not.
		if(tagof(o) == tagof(Treeop.Mouse))
			blkd = isblocked(t.dslash.path, blocked);
		else
			blkd = isblocked(o.path, blocked);
		if(blkd){
			onhold = o::onhold;
			continue Loop;
		}

		elems: list of string;
		elems = nil;
		if(o.path != nil)
			elems = names->elements(o.path);
		if(elems != nil)
			elems = tl elems;	# get rid of "/"

		# 2. Walk to panel
		#
		rp: ref Panel;
		newname: string;	# name for a new panel or nil
		newname = nil;

		# In general, walktree blocks while a subtree is busy
		# but Mouse, Kbd, and Ptwalk try to be optimistic, and
		# only block while the particular panel is busy (due to itself
		# or due to an ancestor). But operation would proceed when
		# if an inner panel is marked as busy.
		pick op := o {
		Mouse =>
			rp = ptwalktree(t.dslash, op.m.xy, 1);
			if(rp != nil)
				t.focus = rp;
		Kbd =>
			if(t.focus == nil)
				t.focus = ptwalktree(t.dslash, lastxy(), 1);
			rp = t.focus;
		Ptwalk =>
			rp = ptwalktree(t.dslash, op.pt, op.atomok);
		Update =>
			rp = walktree(t.slash, elems);
			if(rp == nil){
				# must be new panel, walk to parent instead.
				ppath := dirname(o.path);
				newname = basename(o.path, nil);
				pelems := names->elements(ppath);
				rp = walktree(t.slash, pelems);
			}
			if(rp == nil){
				if(debug['P'])
					fprint(stderr, "walktree: %s: no panel\n", o.path);
				o.rc <-= nil;
				continue Loop;
			}
		* =>
			rp = walktree(t.slash, elems);
			if(rp == nil){
				if(debug['P'])
					fprint(stderr, "walktree: %s: no panel\n", o.path);
				o.rc <-= nil;
				continue Loop;
			}
		}
		if(rp == nil){			# Mouse, Kbd, or Ptwalk
			onhold = o::onhold;		# got their panel busy.
			continue Loop;		# Try later.
		}
		if(rp != nil && (rp.flags&Pdead)){
			o.rc <-= nil;
			continue Loop;
		}

		# 3. Perform the requested operation
		#
		pick op := o {
		Ptwalk =>
			if(!op.atomok)
				rp = paneltop(rp);
			o.rc <-= rp;
		Mouse =>
			t.focus = rp;
			# Mouse on tags and columns may lead to tree operations.
			# we can't set the involved panels busy or we'll deadlock.
			# For containers, the panel is set busy by hand (not using
			# the arg in Xop) and tagmouse is responsible for notifying
			# t.donec once we know it's safe to operate on the panel
			# (i.e., when a menu is no longer shown).
			if(rp.rowcol != Qatom || ((rp.flags&Ptag) && intag(rp, op.m.xy))){
				rp.flags |= Pbusy;
				blocked = rp.path::blocked;
				xc <-= ref Xop(mousetreeproc, t, rp, op, nil, o.rc, nil);
			} else {
				rp.flags |= Pbusy;
				blocked = rp.path::blocked;
				x := ref Xop(mousetreeproc, t, rp, op, t.donec, o.rc, nil);
				xc <-= x;
			}
		Kbd =>
			if(rp.rowcol == Qatom){
				rp.flags |= Pbusy;
				blocked = rp.path::blocked;
				xc <-= ref Xop(kbdtreeproc, t, rp, op, t.donec, o.rc, nil);
			} else {
				xc <-= ref Xop(kbdtreeproc, t, rp, op, nil, nil, nil);
				o.rc <-= rp;
			}
		Walk =>
			o.rc <-= rp;
		Update =>
			updatetree(rp, newname, op.vers, op.ctls, op.data, op.edits);
			o.rc <-= rp;
		Close =>
			if(isprefix(rp.path, t.dslash.path))
				terminate("Root panel removed");
			closetree(rp);
			o.rc <-= rp;
		Layout =>
			force := t.dslash != rp;
			t.dslash = rp;
			rp.flags |= Pbusy;
			blocked = rp.path::blocked;
			notshown(t.slash, t.dslash);
			layoutm->layout(rp, op.auto);
			# kludge: nil tree in the Xop means force
			tt := t;
			if(force)
				tt = nil;
			xc <-= ref Xop(showtreeproc, tt, rp, op, t.donec, o.rc, nil);
		Size =>
			sizetree(op, rp);
			o.rc <-= rp;
		Tags =>
			tagtree(rp, 0);
			o.rc <-= rp;
		Dump =>
			dumptree(rp, 0);
			o.rc <-= rp;
		}
	}
}

Tree.start(name: string): ref Tree
{
	slash := Panel.new(name, nil);
	if(slash == nil){
		werrstr("bad panel type");
		return nil;
	}
	slash.path = "/" + name;
	opc := chan of ref Treeop;
	donec := getpchan();
	if(debug['P'])
		fprint(stderr, "tree slash %s\n", slash.path);
	t := ref Tree("/" + name, slash, slash, opc, donec, nil);
	spawn treeproc(t);
	return t;
}

Tree.terminate(t: self ref Tree)
{
	t.opc <-= nil;
}

# This is called from a mouse op, which means all the tree is busy
# and there should be no race to access the parent.
# The resize tries to honor the target set by the user, but only by
# assigning a %, so that layout might later decide otherwise to honor
# minimum and fixed panel sizes.
resize(nil: ref Tree, p: ref Panel, m: ref Cpointer, mc: chan of ref Cpointer)
{
	setcursor(Resize);
	b := m.buttons;
	xy := m.xy;
	while(m.buttons == b)
		m = <-mc;
	if(m.buttons != 0){
		do {
			m = <-mc;
		} while(m.buttons != 0);
		setcursor(Arrow);
		return;
	}
	setcursor(Arrow);
	r := Rect(xy, m.xy);
	d := Point(r.dx(), r.dy());
	if(xy.x < (p.rect.min.x + p.rect.max.x)/2)
		d.x = -d.x;
	if(xy.y < (p.rect.min.y + p.rect.max.y)/2)
		d.y = -d.y;
	sz := Point(p.rect.dx(), p.rect.dy());
	sz = sz.add(d);
	dp := p.parent;
	if(dp == nil || !(dp.flags&Pshown))
		return;
	dpsz := Point(dp.rect.dx(), dp.rect.dy());
	npcent := Point(sz.x*100/dpsz.x, sz.y*100/dpsz.y);
	if(p.rszpcent.x == 0)
		p.rszpcent.x = p.rect.dx()*100/dpsz.x;
	if(p.rszpcent.y == 0)
		p.rszpcent.y = p.rect.dy()*100/dpsz.y;
	dpcent := npcent.sub(p.rszpcent);
	tot := 0;
	for(i := 0; i < len dp.child; i++){
		np := dp.child[i];
		if((np.flags&Phide) || np == p)
			continue;
		if(dp.rowcol == Qrow && np.wants.x){
			if(np.rszpcent.x == 0)
				np.rszpcent.x = np.rect.dx() * 100 / dpsz.x;
			tot += np.rszpcent.x;
		}
		if(dp.rowcol == Qcol && np.wants.y){
			if(np.rszpcent.y == 0)
				np.rszpcent.y = np.rect.dy() * 100 / dpsz.y;
			tot += np.rszpcent.y;
		}
	}
	if(tot == 0)
		return;
		
	for(i = 0; i < len dp.child; i++){
		np := dp.child[i];
		if((np.flags&Phide) || np == p)
			continue;
		if(dp.rowcol == Qrow && np.wants.x)
			np.rszpcent.x -= dpcent.x * np.rszpcent.x / tot;
		if(dp.rowcol == Qcol && np.wants.y)
			np.rszpcent.y -= dpcent.y * np.rszpcent.y / tot;
		if(np.rszpcent.x <= 0)
			np.rszpcent.x = 0;
		if(np.rszpcent.y <= 0)
			np.rszpcent.y = 0;
	}
	p.rszpcent = npcent;
}

insertpoint(p: ref Panel, pt: Point): int
{
	pos := i := 0;
	if(p.rowcol == Qrow)
		for(i = 0; i < len p.child; i++){
			np := p.child[i];
			if(np.flags&Pshown)
			if(pt.x > np.rect.min.x + np.rect.dx()/2)
				pos = i+1;
			else if(pt.x < np.rect.min.x + np.rect.dx()/2)
				break;
		}
	else
		for(i = 0; i < len p.child; i++){
			np := p.child[i];
			if(np.flags&Pshown)
			if(pt.y > np.rect.min.y + np.rect.dy()/2)
				pos = i+1;
			else if(pt.y < np.rect.min.y + np.rect.dy()/2)
				break;
		}
	return pos;
}

drag(t: ref Tree, p: ref Panel, m: ref Cpointer, mc: chan of ref Cpointer, innerok: int)
{
	setcursor(Drag);
	b := m.buttons;
	xy := m.xy;
	while(m.buttons == b){
		m = <-mc;
	}
	if(m.buttons != 0){
		do {
			m = <-mc;
		} while(m.buttons != 0);
		setcursor(Arrow);
		return;
	}
	r := Rect(xy, m.xy);
	r = r.canon();
	if(p.rowcol && r.dx() < 10 && p.rect.contains(m.xy)){
		p.ctl("col");
		p.fsctl("col", 1);
		t.layout(nil, 0);
	} else if(p.rowcol && r.dy() < 10 && p.rect.contains(m.xy)){
		p.ctl("row");
		p.fsctl("row", 1);
		t.layout(nil, 0);
	} else {
		np := t.ptwalk(m.xy, innerok);
		if(innerok && np.rowcol == Qatom)
			np = np.parent;
		if(np == p){
			pos := insertpoint(np, m.xy);
			p.fsctl(sprint("pos %d\n", pos), 1);
		} else if(np != nil && np != p){
			pos := insertpoint(np, m.xy);
			path := np.path;
			if(copying)
				p.fsctl(sprint("copyto %s %d", path, pos), 1);
			else
				p.fsctl(sprint("moveto %s %d", path, pos), 1);
			copying = 0;
		}
	}
	setcursor(Arrow);
}

copying := 0;

wcmd(t: ref Tree, p: ref Panel, c: string)
{
	if(c != "Close")
		p.flags &= ~Pinsist;
	case c {
	"Copy" =>
		copying = 1;
		pt := p.rect.min;
		pt.add((Inset+Tagwid/2, Inset+Taght/2));
		win.wmctl("ptr " + string pt.x + " " + string pt.y);
	"More" =>
		t.size(t.path(p), More);
		t.tags("/");
		t.layout(nil, 0);
	"Hide" =>
		p.ctl("hide");
		p.fsctl("hide", 1);
		t.tags("/");
		t.layout(nil, 0);
	"Close" =>
		if(p.flags&(Pdirties|Pdirty))
			if(!(p.flags&Pinsist)){
				p.flags |= Pinsist; 
				p.fsctl("exec !echo dirty panels", 1);
				return;
			}
		if(t.slash != p && t.dslash != p)
			p.fsctl("exec Close", 1);
		else
			terminate(nil);
	"Top" =>
		if(p == t.dslash)
			t.layout("/", 0);
		else
			t.layout(t.path(p), 0);
		pt := p.rect.min.add((Tagwid/2, Taght/2));
		win.wmctl("ptr " + string pt.x + " " + string pt.y);
	"Full" =>
		t.size(t.path(p), Full);
		t.tags("/");
		t.layout(nil, 0);
	* =>
		fprint(stderr, "o/live: wcmd: unknown cmd %s\n", c);
	}
	if(debug['P'])
		t.dump("/");
}

tagmenu: ref Menu;

tagmouse1(t: ref Tree, p: ref Panel, m: ref Cpointer, mc: chan of ref Cpointer)
{
	m = <-mc;
	case m.buttons {
	1 =>
		setcursor(Resize);
		m = <-mc;
		case m.buttons {
		0 =>
			t.layout(nil, 1);
			setcursor(Arrow);
			return;
		1 =>
			resize(t, p, m, mc);
			t.layout(nil, 0);
			setcursor(Arrow);
			return;
		}
	}
	setcursor(Arrow);	
	while(m.buttons)
		m = <-mc;
}

# We're may be called from rowcol's pmouse(), in that case, the panel is
# busy only until we know it's not a menu. For atoms, the panel is busy as usual
# and the tree will release the hold once we return.
# That means that we must send to donec for rows and columns.
tagmouse(t: ref Tree, p: ref Panel, m: ref Cpointer, mc: chan of ref Cpointer)
{
	if(tagmenu == nil)
		tagmenu = Menu.new(array[] of
			{"Copy", "More", "Hide", "Close", "Top", "Full"});
	case m.buttons {
	1 =>
		if(p.rowcol != Qatom);
			t.donec <-= p;
		m = <-mc;
		case m.buttons {
		1 =>
			drag(t, p, m, mc, 0);
		0 =>
			tagmouse1(t, p, m, mc);
		* =>
			while(m.buttons)
				m = <-mc;
		}
	2 =>
		if(p.rowcol != Qatom)
			t.donec <-= p;
		if(cookclick(m, mc)){
			t.size(t.path(p), Max);
			t.tags("/");
			t.layout(nil, 0);
		}
	4 =>
		m = <-mc;
		case m.buttons {
		0 =>
			copying = 0;
			opt := tagmenu.run(m, mc);
			if(p.rowcol != Qatom)
				t.donec <-= p;
			if(opt != nil)
				wcmd(t, p, opt);
		* =>
			t.donec <-= p;
			while(m.buttons)
				m = <-mc;
		}
	* =>
		if(p.rowcol != Qatom)
			t.donec <-= p;
		while(m.buttons)
			m = <-mc;
	}
}

panelmouse(t: ref Tree, p: ref Panel, m: ref Cpointer, mc: chan of ref Cpointer): int
{
	if((p.flags&Ptag) && intag(p, m.xy)){
		tagmouse(t, p, m, mc);
		return 1;
	}
	case m.buttons {
	2 =>
		if(cookclick(m, mc))
			p.fsctl(sprint("exec %s", p.name), 1);
		return 1;
	4 =>
		if(cookclick(m, mc))
			p.fsctl(sprint("look %s", p.name), 1);
		return 1;
	* =>
		# let the panel handle other events.
		return 0;
	}
}

panelkbd(nil: ref Tree, p: ref Panel, r: int)
{
	if(r == Keyboard->Del)
		p.fsctl("interrupt", 1);
	else if(p.flags&Pedit)
		p.fsctl(sprint("keys %c", r), 1);
}

flagredraw(fp: ref Panel)
{
	fp.flags |= Predraw;
	for(i := 0; i < len fp.child; i++)
		flagredraw(fp.child[i]);
}

flaghide(fp: ref Panel)
{
	fp.flags &= ~Pshown;
	fp.rect = Rect((0,0),(0,0));	# just be sure of it.
	for(i := 0; i < len fp.child; i++)
		flaghide(fp.child[i]);
}

# The FS should never report certain attributes for certain panels.
# In any case, we process most of them here, and most panels may
# call this function to deal with their ctls.
panelctl(t: ref Tree, p: ref Panel, s: string): int
{
	(nargs, args) := tokenize(s, " \t\n");
	if(nargs < 1)
		return -1;
	case hd args {
	"row" =>
		if(p.rowcol != Qrow)
			flagredraw(p);
		p.rowcol = Qrow;
	"col" =>
		if(p.rowcol != Qcol)
			flagredraw(p);
		p.rowcol = Qcol;
	"appl" =>
		# arg ignored
		p.flags &= ~Playout;
	"layout" =>
		p.flags |= Playout;
	"hide" =>
		p.flags |= Phide;
		flaghide(p);
	"show" =>
		p.flags &= ~Phide;
		flagredraw(p);
	"tag" =>
		if(!(p.flags&Ptag))
			p.flags |= Predraw;
		p.flags |= Ptag;
	"notag" =>
		if(p.flags&Ptag)
			p.flags |= Predraw;
		p.flags &= ~Ptag;
	"dirty" =>
		# containers should never get here
		if(!(p.flags&Pdirty))
			p.flags |= Predraw;
		p.flags |= Pdirty;
	"clean" =>
		# containers should never get here
		if(p.flags&Pdirty)
			p.flags |= Predraw;
		p.flags &= ~Pdirty;
	"font" =>
		o := p.font;
		# only text panels should get here
		p.font = getfont(nth(args, 1));
		if(o != p.font)
			p.flags |= Predraw;
	"order" =>
		if(tl args != p.order){
			p.order = tl args;
			orderchildren(p);
			flagredraw(p);
		}
	"debug" =>
		flgs := hd tl args;
		p.setdebug(0, 0);
		i: int;
		if(p == t.slash)
			for(i = 0; i < len debug; i++)
				debug[i] = 0;
		for(i = 0; i < len flgs; i++)
			if(p == t.slash)
				debug[flgs[i]] = 1;
			else
				p.setdebug(flgs[i], 1);
	* =>
		return -1;
	}
	return 0;
}
